import { __awaiter, __generator, __values } from "tslib";
import React from "react";
import { createRoot } from "react-dom/client";
import { Provider } from "react-redux";
import { Router } from "react-router-dom";
import axios from "axios";
import { NavigationGuard } from "./NavigationGuard";
import { app } from "../app";
import { executeAction } from "../module";
import { idleTimeoutAction } from "../reducer";
import { APIException } from "../Exception";
import { call, delay } from "../typed-saga";
import { ErrorBoundary } from "../util/ErrorBoundary";
import { ajax } from "../util/network";
import { isBrowserSupported, isIOS } from "../util/navigator-util";
import { captureError, errorToException } from "../util/error-util";
import { DEFAULT_IDLE_TIMEOUT, IdleDetector } from "../util/IdleDetector";
export var LOGGER_ACTION = "@@framework/logger";
export var VERSION_CHECK_ACTION = "@@framework/version-check";
export var GLOBAL_ERROR_ACTION = "@@framework/global";
export var GLOBAL_PROMISE_REJECTION_ACTION = "@@framework/promise-rejection";
export function bootstrap(option) {
    var _a, _b, _c, _d, _e;
    detectOldBrowser((_a = option.browserConfig) === null || _a === void 0 ? void 0 : _a.onOldBrowserDetected);
    setupGlobalErrorHandler(option.errorListener);
    setupAppExitListener((_b = option.loggerConfig) === null || _b === void 0 ? void 0 : _b.serverURL);
    setupLocationChangeListener((_c = option.browserConfig) === null || _c === void 0 ? void 0 : _c.onLocationChange);
    setupIdleTimeout((_d = option.idleTimeoutInSecond) !== null && _d !== void 0 ? _d : DEFAULT_IDLE_TIMEOUT);
    runBackgroundLoop(option.loggerConfig, option.versionConfig);
    renderRoot(option.componentType, option.rootContainer || injectRootContainer(), ((_e = option.browserConfig) === null || _e === void 0 ? void 0 : _e.navigationPreventionMessage) || "Are you sure to leave current page?");
}
function detectOldBrowser(onOldBrowserDetected) {
    if (!isBrowserSupported()) {
        if (onOldBrowserDetected) {
            onOldBrowserDetected();
        }
        else {
            var alertMessage = void 0;
            var navigatorLanguage = navigator.languages ? navigator.languages[0] : navigator.language;
            if (navigatorLanguage.startsWith("zh")) {
                alertMessage = "对不起，您的浏览器版本太老不支持\n请使用最新版的浏览器再访问";
            }
            else {
                alertMessage = "Your browser is not supported.\nPlease upgrade your browser to latest and visit again.\nSorry for the inconvenience.";
            }
            alert(alertMessage);
        }
        // After that, the following code may still run
    }
}
function setupGlobalErrorHandler(errorListener) {
    app.errorHandler = errorListener.onError.bind(errorListener);
    window.addEventListener("error", function (event) {
        try {
            var analyzeByTarget = function () {
                if (event.target && event.target !== window) {
                    var element = event.target;
                    return "DOM source error: ".concat(element.outerHTML);
                }
                return "Unrecognized error, serialized as ".concat(JSON.stringify(event));
            };
            captureError(event.error || event.message || analyzeByTarget(), GLOBAL_ERROR_ACTION);
        }
        catch (e) {
            /**
             * This should not happen normally.
             * However, global error handler might catch external webpage errors, and fail to parse error due to cross-origin limitations.
             * A typical example is: Permission denied to access property `foo`
             */
            app.logger.warn({
                action: GLOBAL_ERROR_ACTION,
                errorCode: "ERROR_HANDLER_FAILURE",
                errorMessage: errorToException(e).message,
                elapsedTime: 0,
                info: {},
            });
        }
    }, true);
    window.addEventListener("unhandledrejection", function (event) {
        try {
            captureError(event.reason, GLOBAL_PROMISE_REJECTION_ACTION);
        }
        catch (e) {
            app.logger.warn({
                action: GLOBAL_PROMISE_REJECTION_ACTION,
                errorCode: "ERROR_HANDLER_FAILURE",
                errorMessage: errorToException(e).message,
                elapsedTime: 0,
                info: {},
            });
        }
    }, true);
}
function renderRoot(EntryComponent, rootContainer, navigationPreventionMessage) {
    var root = createRoot(rootContainer);
    root.render(React.createElement(Provider, { store: app.store },
        React.createElement(IdleDetector, null,
            React.createElement(Router, { history: app.history },
                React.createElement(NavigationGuard, { message: navigationPreventionMessage }),
                React.createElement(ErrorBoundary, null,
                    React.createElement(EntryComponent, null))))));
}
function injectRootContainer() {
    var rootContainer = document.createElement("div");
    rootContainer.id = "framework-app-root";
    document.body.appendChild(rootContainer);
    return rootContainer;
}
function setupAppExitListener(eventServerURL) {
    if (eventServerURL) {
        // ref: https://developer.apple.com/library/archive/documentation/AppleApplications/Reference/SafariWebContent/HandlingEvents/HandlingEvents.html#//apple_ref/doc/uid/TP40006511-SW5
        window.addEventListener(isIOS() ? "pagehide" : "unload", function () {
            try {
                app.logger.info({ action: "@@EXIT" });
                var logs = app.logger.collect();
                /**
                 * navigator.sendBeacon() uses HTTP POST, but does not support CORS.
                 * We have to use text/plain as content type, instead of JSON.
                 */
                var textData = JSON.stringify({ events: logs });
                navigator.sendBeacon(eventServerURL, textData);
            }
            catch (e) {
                // Silent if sending error
            }
        }, false);
    }
}
function setupLocationChangeListener(listener) {
    if (listener) {
        app.history.listen(listener);
    }
}
function setupIdleTimeout(timeout) {
    app.store.dispatch(idleTimeoutAction(timeout));
}
function runBackgroundLoop(loggerConfig, versionCheckConfig) {
    app.logger.info({ action: "@@ENTER" });
    app.loggerConfig = loggerConfig || null;
    if (loggerConfig) {
        app.sagaMiddleware.run(function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!true) return [3 /*break*/, 3];
                        return [4 /*yield*/, delay((loggerConfig.frequencyInSecond || 20) * 1000)];
                    case 1:
                        _a.sent();
                        return [5 /*yield**/, __values(call(sendEventLogs))];
                    case 2:
                        _a.sent();
                        return [3 /*break*/, 0];
                    case 3: return [2 /*return*/];
                }
            });
        });
    }
    if (versionCheckConfig) {
        app.sagaMiddleware.run(function () {
            var lastChecksum, newChecksum;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        lastChecksum = null;
                        _a.label = 1;
                    case 1:
                        if (!true) return [3 /*break*/, 7];
                        return [5 /*yield**/, __values(call(fetchVersionChecksum, versionCheckConfig.versionCheckURL))];
                    case 2:
                        newChecksum = _a.sent();
                        if (!newChecksum) return [3 /*break*/, 5];
                        if (!(lastChecksum !== null && newChecksum !== lastChecksum)) return [3 /*break*/, 4];
                        return [5 /*yield**/, __values(executeAction(VERSION_CHECK_ACTION, versionCheckConfig.onRemind))];
                    case 3:
                        _a.sent();
                        _a.label = 4;
                    case 4:
                        lastChecksum = newChecksum;
                        _a.label = 5;
                    case 5: return [4 /*yield*/, delay((versionCheckConfig.frequencyInSecond || 600) * 1000)];
                    case 6:
                        _a.sent();
                        return [3 /*break*/, 1];
                    case 7: return [2 /*return*/];
                }
            });
        });
    }
}
export function sendEventLogs() {
    return __awaiter(this, void 0, void 0, function () {
        var logs, logLength, e_1;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (!app.loggerConfig) return [3 /*break*/, 4];
                    logs = app.logger.collect(200);
                    logLength = logs.length;
                    if (!(logLength > 0)) return [3 /*break*/, 4];
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    /**
                     * Event server URL may be different from current domain (supposing abc.com)
                     *
                     * In order to support this, we must ensure:
                     * - Event server allows cross-origin request from current domain
                     * - Root-domain cookies, whose domain is set by current domain as ".abc.com", can be sent (withCredentials = true)
                     */
                    return [4 /*yield*/, ajax("POST", app.loggerConfig.serverURL, {}, { events: logs }, { withCredentials: true })];
                case 2:
                    /**
                     * Event server URL may be different from current domain (supposing abc.com)
                     *
                     * In order to support this, we must ensure:
                     * - Event server allows cross-origin request from current domain
                     * - Root-domain cookies, whose domain is set by current domain as ".abc.com", can be sent (withCredentials = true)
                     */
                    _a.sent();
                    app.logger.emptyLastCollection();
                    return [3 /*break*/, 4];
                case 3:
                    e_1 = _a.sent();
                    if (e_1 instanceof APIException) {
                        // For APIException, retry always leads to same error, so have to give up
                        // Do not log network exceptions
                        app.logger.emptyLastCollection();
                        app.logger.exception(e_1, {
                            action: LOGGER_ACTION,
                            info: { dropped_logs: logLength.toString() },
                        });
                    }
                    return [3 /*break*/, 4];
                case 4: return [2 /*return*/];
            }
        });
    });
}
/**
 * Only call this function if necessary, i.e: initial checksum, or after long-staying check
 * Return latest checksum, or null for failure.
 */
function fetchVersionChecksum(url) {
    return __awaiter(this, void 0, void 0, function () {
        var startTime, response, checksum, e_2;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    startTime = Date.now();
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, axios.get(url)];
                case 2:
                    response = _a.sent();
                    checksum = JSON.stringify(response.data);
                    app.logger.info({
                        action: VERSION_CHECK_ACTION,
                        elapsedTime: Date.now() - startTime,
                        info: { checksum: checksum },
                    });
                    return [2 /*return*/, checksum];
                case 3:
                    e_2 = _a.sent();
                    app.logger.warn({
                        action: VERSION_CHECK_ACTION,
                        elapsedTime: Date.now() - startTime,
                        errorCode: "VERSION_CHECK_FAILURE",
                        errorMessage: errorToException(e_2).message,
                    });
                    return [2 /*return*/, null];
                case 4: return [2 /*return*/];
            }
        });
    });
}
//# sourceMappingURL=bootstrap.js.map